// Header File
// Created 3/25/2004; 9:05:29 AM

//-----------------------------------------------------------------------------
// structure of a process context which holds necessary informations of a task
// (registers, some tios variables, flags ...)
//-----------------------------------------------------------------------------
typedef struct _PROCESSCONTEXT {
    unsigned long           registers[15];  // holds d0-d7 & a0-a6
    unsigned long           usp;            // a7
    unsigned short          sr;             // state register
    unsigned long           pc;             // program counter
    unsigned long           tiosdata[9];    // some important tios variables
    struct _PROCESSCONTEXT* next;
    unsigned short          id;
    

    union {
        unsigned short all;
        struct {
            unsigned int flag1_7 : 1;
            unsigned int flag1_6 : 1;
            unsigned int flag1_5 : 1;
            unsigned int flag1_4 : 1;
            unsigned int flag1_3 : 1;
            unsigned int flag1_2 : 1;
            unsigned int flag1_1 : 1;
            unsigned int flag1_0 : 1;

            unsigned int flag0_7 : 1;
            unsigned int flag0_6 : 1;
            unsigned int flag0_5 : 1;
            unsigned int flag0_4 : 1;
            unsigned int flag0_3 : 1;
            unsigned int flag0_2 : 1;
            unsigned int flag0_1 : 1;
            unsigned int active  : 1;
        } bits;
    } flags;

} PROCESSCONTEXT;


//-----------------------------------------------------------------------------
// return values function EnableMultitasking()
//-----------------------------------------------------------------------------
#define MT_OK              0   // enabled correctly
#define MT_ALREADY_RUNNING 1   // multitasking already running - nothing done


//-----------------------------------------------------------------------------
// ONLY the following functions should be used directly (!!!)
//-----------------------------------------------------------------------------
short EnableMultitasking(unsigned long handler_index);
void  DisableMultitasking(void);
void  StartProcess(PROCESSCONTEXT* context,void* entrypoint,void* stack,unsigned short stacksize);


//#############################################################################
//#############################################################################
//
// INTERNAL IMPLEMENTATION
//
//#############################################################################
//#############################################################################

//-----------------------------------------------------------------------------
// internally used globals
//-----------------------------------------------------------------------------
static volatile PROCESSCONTEXT* _current_process_ = 0; // process list
static volatile PROCESSCONTEXT  _default_process_;     // the default process
static volatile unsigned long*  _used_handler_;        // handler index (i.e. AUTO_INT1 ...)
static volatile unsigned long   _old_trap13_;          // old trap 13 handler pointer
extern unsigned long            _old_int1_;            // old trap 1 handler pointer


void _Scheduler(void); // forward declaration of ASM task scheduler



//=============================================================================
// fetches the next process in the linked list which is active.
//
// NOTE: the end of the linked list haven't to link to NULL, but back to the
// start of the list!
//
// INTERNAL function - should only be called from Scheduler
//=============================================================================

// note that this function isn't declared static, because it is used by the
// ASM scheduler and the compiler doesn't know this. If it is declared static
// the compiler will output a warning.
PROCESSCONTEXT* _GetNextProcess(void) {
    PROCESSCONTEXT* tmp = _current_process_->next;

    while (!(tmp->flags.bits.active)) tmp = tmp->next;
    return tmp;
}


//=============================================================================
// The Context switch interrupt: save the current process context, call the
// scheduler to select what process is next to run, restore its context, and
// give it the CPU.
//
// INTERNAL function - DON'T CALL IT directly
//=============================================================================
asm("_Scheduler:\n"\
"       or.w        #0x0700,%sr\n"\
"       move.l      %a6,-(%sp)\n"\
"       move.l      _current_process_,%a6\n"\
"       movem.l     %d0-%d7/%a0-%a5,0(%a6)      | store d0-d7 & a0-a5 in process context\n"\
"       move.l      (%sp)+,14*4(%a6)            | store a6\n"\
"       move.l      %usp,%a1\n"\
"       move.l      %a1,15*4(%a6)               | store a7\n"\
"       move.w      (%sp)+,16*4(%a6)            | store sr\n"\
"       move.l      (%sp)+,16*4+2(%a6)          | store pc\n"\
"PATCH1:movem.l     (0x1234),%a0/%d0-%d7\n"\
"       movem.l     %a0/%d0-%d7,16*4+6(%a6)     | store TIOS variables!\n"\
"       jsr         _GetNextProcess             | returns pointer to next process context in %a0\n"\
"       move.l      %a0,_current_process_\n"\
"       movem.l     16*4+6(%a0),%a1/%d0-%d7\n"\
"PATCH2:movem.l     %a1/%d0-%d7,(0x1234)\n"\
"       move.l      15*4(%a0),%a1\n"\
"       move.l      %a1,%usp\n"\
"       move.l      16*4+2(%a0),-(%sp)\n"\
"       move.w      16*4(%a0),-(%sp)\n"\
"       movem.l     0(%a0),%a0-%a6/%d0-%d7\n"\
"       .word    0x4ef9                     | JMP address opcode\n"\
"_old_int1_: .long    0x00000000              | address of old int1 gots stored here");



//=============================================================================
// returns a unique process id
//
// INTERNAL function - DON'T CALL IT directly
//=============================================================================
static unsigned short GetNextID() {
    static unsigned short id_counter = 1;
    return id_counter++;
}


//=============================================================================
// terminates a process by deleting the active flag in the process context
// structure
//
// INTERNAL function - DON'T CALL IT directly
//=============================================================================
static void TerminateProcess(HANDLE pid) {
    PROCESSCONTEXT* tmp = (PROCESSCONTEXT*)_current_process_;

    // NOTE: yet TerminateProcess() is only executed by the process which
    //       should terminate itself. So it doesn't make sense to search
    //       through the process list to fit itself (it IS the _current_process).
    //       But to be aware of future extensions like terminating a process
    //       from another process we will scan the process list for the correct
    //       pid.

    while (tmp->id != pid) tmp = tmp->next;

    tmp->flags.bits.active = 0;

    // if the current process is the process to terminate call the scheduler
    // manually to trigger a process switch
    if (_current_process_ == tmp) asm("trap #13");
}


//=============================================================================
// starts a new process
//
// context ... pointer to PROCESSCONTEXT structure which will be used for this
//             process (MUST be valid UNTIL multitasking gets disabled!)
//
// entrypoint ... pointer to start of new process, normally this is a pointer
//                to a "void f(void)" function
//
// stack      ... pointer to buffer used as stack for this process
// stacksize  ... size of stack
//=============================================================================
void StartProcess(PROCESSCONTEXT* context,void* entrypoint,void* stack,unsigned short stacksize) {
    // Initialize a bunch of context stuff
    context->sr                = 0;
    context->pc                = (unsigned long)entrypoint;
    context->usp               = (unsigned long)stack + stacksize - 10; // stack grows in reverse order!
    context->flags.bits.active = 1;
    context->id                = GetNextID();

    asm("movem.l %%a4-%%a5/%%d0/%%d2-%%d7,-(%%a7);"\
        "PATCH3:  movem.l (0x1234),%%a4-%%a5/%%d0/%%d2-%%d7; movem.l %%a4-%%a5/%%d0/%%d2-%%d7,16*4+6(%0);"\
        "movem.l (%%a7)+,%%a4-%%a5/%%d0/%%d2-%%d7" :: "a" (context));

    // push terminate process function pointer and the process ID on stack so
    // it will be executed when the process finishes.
    *((unsigned long *)(context->usp)) = (unsigned long)TerminateProcess;
    *((unsigned short *)((void *)context->usp + 8)) = context->id;

    // add the process to the process list
    context->next = _current_process_->next;
    _current_process_->next = context;       // now the process is active!
}


//=============================================================================
// enables multitasking (returns MT_OK if correctly enabled)
//
// handler_index ... index of interrupt to use (AUTO_INT_1, AUTO_INT_5 ...)
//=============================================================================
short EnableMultitasking(unsigned long handler_index) {
    if (_current_process_) return MT_ALREADY_RUNNING;

    // We'll start by finding one address that shifts between ROM versions (and calcs)
    // and place it in our code here and there...
    // this address is used to save and restore some important TIOS variables!
    asm("move.w 6(%0),%%a0;move.w %%a0,PATCH1+4;move.w %%a0,PATCH2+4;move.w %%a0,PATCH3+4" ::"a" (&PortRestore):"a0");

    _old_trap13_   = *(unsigned long*)0x000000b4;               // backup old trap 13
    _used_handler_ = (unsigned long*)(handler_index | 0x40000); // backup handler index for disable function
    _old_int1_     = *_used_handler_;                           // backup old handler adress

    *(unsigned long*)0x000400b4 = (unsigned long)_Scheduler;    // setup new trap 13


    // initialize the current_process global ...
    _default_process_.id                = GetNextID();
    _default_process_.flags.bits.active = 1;

    // very stupid "i'm my own successor task" setup -> no real multitasking yet
    _current_process_       = &_default_process_;
    _current_process_->next = (PROCESSCONTEXT*)_current_process_;

    // Now call scheduler once to fill all entries of the default_process
    asm("trap #13");

    *(unsigned long*)_used_handler_ = (unsigned long)_Scheduler;    // setup scheduler

    return MT_OK;
}


//=============================================================================
// disables multitasking
//=============================================================================
void DisableMultitasking(void) {
    if (!_current_process_) return;
    *_used_handler_=_old_int1_;                 // restore old INT1 handler
    *(unsigned long*)0x00400b4 = _old_trap13_;  // restore old trap 13
    _current_process_ = 0;
}